/*
 * Copyright (C) 2021 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"
#include "calls.hpp"
#include "features.hpp"
#include "messages.hpp"

#include <moderncli/keyfile.hpp>
#include <moderncli/process.hpp>

#include <list>

using namespace std::literals::string_literals;
using namespace server;

namespace {
const size_t call_entries = SERVER_CALL_LIMIT;
const auto mem_size = sizeof(pbx_call) * call_entries;

struct args_t final {
    args_t() = default;
    args_t(keyfile cfg) : config(std::move(cfg)) {}
    args_t(event_t evt) : event(std::move(evt)) {}
    args_t(pbx_var evt) : ipc(std::move(evt)) {}
    args_t(registry reg, reason_t r) : entry(std::move(reg)), reason(std::move(r)) {}
    args_t(uint64_t id, reason_t r) : call_id(id), reason(std::move(r)) {}
    args_t(uint64_t id, event_t evt, registry reg) : call_id(id), event(std::move(evt)), entry(std::move(reg)) {}
    args_t(uint64_t id, registry reg, unsigned r) : call_id(id), entry(std::move(reg)), rings(r) {}
    args_t(uint64_t id, event_t evt, feature_t f) : call_id(id), event(std::move(evt)), feature(f), entry(std::move(f->origin)) {}

    uint64_t call_id{0};
    keyfile config;
    event_t event;
    feature_t feature{nullptr};
    pbx_var ipc;
    registry entry;
    reason_t reason;
    unsigned rings{0};
};

using actor_t = void (*)(const args_t& args);
using msg_t = std::pair<actor_t, args_t>;
using mode_t = enum Mode {PROXY, ROUTED, RELEASE};
using ringer_t = std::pair<sync_timepoint, uint64_t>;

// each sip endpoint call leg is a segment
using segment_t = struct Segment final {
    Segment() = delete;

    // construct initial call segment from a sip event and registry entry
    Segment(uint64_t call, const event_t& event, const registry& reg) :
    call_id(call), user_id(reg.user_id()), context(event.context()), cid(event.cid()), tid(event.tid()), entry(reg), state(RING) {}

    // construct an added segment from registry that rings at x
    Segment(uint64_t call, const registry& reg, unsigned ring = 0) :
    call_id(call), user_id(reg.user_id()), context(reg.context()), ring_at(ring), entry(reg), state(INVITE) {}

    // terminate an active call segment, could be private...
    void terminate(const char *reason = nullptr) const {
        logger.debug(5, "terminate segment {}/{} dialog {}", cid, user_id, did);
        const exosip::guard_t lock(context);
        if(reason)
            eXosip_call_terminate_with_reason(context, cid, did, reason);
        else
            eXosip_call_terminate(context, cid, did);
    }

    // invite endpoint thru this segment
    auto invite(const char *from, const char *to, const char *subject, std::function<void(exosip::msg_t)> fill = [](auto msg){}) {
        const auto& route = entry.route_to();
        auto msg = exosip::msg_t(nullptr);

        const exosip::guard_t lock(context);
        eXosip_call_build_initial_invite(context, &msg, to, from, route.c_str(), subject);
        if(!msg)
            return -1;

        fill(msg);
        cid = eXosip_call_send_initial_invite(context, msg);
        return cid;
    }

    // send ack to endpoint for this segment
    auto ack(std::function<void(exosip::msg_t)> fill = [](auto msg){}) {
        auto msg = exosip::msg_t(nullptr);
        auto id = tid;
        if(id < 1)
            id = did;
        const exosip::guard_t lock(context);
        eXosip_call_build_ack(context, id, &msg);
        if(!msg)
            return false;

        fill(msg);
        eXosip_call_send_ack(context, id, msg);
        ackflag = true;
        return true;
    }

    // construct and send call answer to endpoint for this segment
    auto answer(int status, std::function<void(exosip::msg_t)> fill = [](auto msg){}) const
    {
        if(tid < 1)
            return false;
        auto msg = exosip::msg_t(nullptr);
        const exosip::guard_t lock(context);
        eXosip_call_build_answer(context, tid, status, &msg);
        if(!msg)
            return false;

        fill(msg);
        eXosip_call_send_answer(context, tid, status, msg);
        return true;
    }

    // close the current segment, update segment state
    void close(int status = 0, const char *reason = nullptr) {
        if(state == Segment::CLOSE || state == Segment::DELAY)
            return;

        if(!active || state != Segment::RING)
            tid = -1;

        active = false;
        state = Segment::CLOSE;
        --entry.shm()->count;   // TODO: Maybe move to release???

        if(!status) {
            logger.debug(5, "ending segment {}/{}, skipping", cid, user_id);
            return;
        }

        if(tid > 0) {
            logger.debug(5, "ending segment {}/{}, status={}, active={}", cid, user_id, status, active);
            answer(status);
            tid = -1;
            return;
        }

        logger.debug(5, "ending segment {}/{}, terminate", cid, user_id);
        terminate(reason);
    }

    // pod data...
    uint64_t call_id;           // can find call from segment...
    std::string user_id;
    eXosip_t* context;
    int cid{-1};
    int tid{-1};
    int did{-1};
    unsigned ring_at{0};        // when to start ringing...
    registry entry;
    enum {INVITE, RING, JOINED, CLOSE, DELAY} state;
    bool active{false};         // "active" segment...
    bool ackflag{false};        // processed ack...
    bool reinvited{false};      // re-invited...
};

// a sip call that interconnects segments
using call_t =  struct Call final {
    Call() = delete;

    // construct a sip call from global id and shm block
    Call(uint64_t call_id, pbx_call *mem) : shm(mem) {
        time(&shm->created);
        shm->id = call_id;
        shm->active = 1;
        shm->ringing = 1;
        shm->segment = 0;
        shm->state = pbx_call::INVITING;
    };

    // apply lambda to active segments not matching this cid
    auto apply(std::function<bool(segment_t&)> pred, int cid = 0) {
        auto count = 0U;
        for(auto& segment : segments) {
            if(segment.active && (!cid || segment.cid != cid) && pred(segment))
                ++count;
        }
        return count;
    }

    // determine reason a call is closing based on what happened
    auto close_status() const {
        if(!declines && busy)
            return SIP_BUSY_HERE;
        return SIP_DECLINE;
    }

    pbx_call *shm;
    std::list<segment_t> segments;
    std::string from, to, subject, sdp;
    unsigned busy{0}, declines{0}, rings{0};
    time_t joined{0};
    time_t expires{0};      // expires from invite
    time_t period{0};       // period interval active
    time_t minimal{0};      // minimal acceptable period if timed
};

std::atomic<unsigned> sequence(0);
std::atomic<unsigned> active_count{0};
mode_t mode = Mode::PROXY;
auto ring = 4;
auto delayed = 3U;
auto stopping = false;

std::condition_variable cvar;
unique_sync<std::list<msg_t>> msglist;

keyfile::keys display_names;
std::unordered_map<uint64_t, call_t> calls;
std::unordered_map<int, segment_t*> sessions;     // active segments
std::list<pbx_call*> freeipc;
pbx_call *mapped{};

// Ringer is always in time order because new ringers are always incremented by
// the same amount of time, and ringers are created at the back o the list.
// The oldest ringer is always at the front. Ringer is only manipulated from
// the exec_calls thread and exec_calls initiated functions.
std::list<ringer_t> ringer;

// our helpers

void cdr_logger(call_t& call, int code) {
    if(!call.joined)
        return;

    auto type = "?";
    switch(call.shm->type) {
    case pbx_call::LOCAL:
        type = "local";
        break;
    case pbx_call::INBOUND:
        type = "public";
        break;
    case pbx_call::OUTGOING:
        type = "remote";
        break;
    default:
        break;
    }

    auto msg = format( R"("id":"0x{:016x}","sn":{},"rt":"{}","ci":"{}","di":"{}","rh":"{}","xc":{})",
        call.shm->id, ++call.shm->segment, type, call.shm->caller, call.shm->dialed, call.shm->remote, code);
    exec::post_cdr(logmsg(msg, call.joined));
    call.joined = 0;
}

void reply_event(const event_t& event, int status) {
    auto context = event.context();
    auto tid = event.tid();
    osip_message_t *msg = nullptr;
    const exosip::guard_t lock(context);
    eXosip_call_build_answer(context, tid, status, &msg);
    eXosip_call_send_answer(context, tid, status, msg);
}

void send_info(eXosip_t *context, int did, const char *type, osip_body_t *body) {
    osip_message_t *msg = nullptr;
    const exosip::guard_t lock(context);
    if(eXosip_call_build_info(context, did, &msg) != 0)
        return;

    osip_message_set_body(msg, body->body, body->length);
    osip_message_set_content_type(msg, type);
    eXosip_call_send_request(context, did, msg);
}

void send_invite(eXosip_t *context, int did, const char *type, osip_body_t *body) {
    osip_message_t *msg = nullptr;
    const exosip::guard_t lock(context);
    if(eXosip_call_build_request(context, did, "INVITE", &msg) != 0)
        return;

    if(body && body->body && body->length) {
        osip_message_set_body(msg, body->body, body->length);
        osip_message_set_content_type(msg, type);
    }

    eXosip_call_send_request(context, did, msg);
}

// our ringers

void create_ringer(uint64_t call_id) {
    auto now = sync_clock();
    now += std::chrono::seconds(ring);
    ringer.emplace_back(now, call_id);
}

auto delayed_ringer(call_t& call, unsigned rings) {
    auto sdp = call.sdp;              // originators sdp
    auto from = call.from.c_str();
    auto to = call.to.c_str();
    auto subject = call.subject.c_str();
    auto count = 0;

    // invite delayed ring segments
    for(auto& segment : call.segments) {
        if(segment.cid != -1 || segment.ring_at > rings || segment.state != Segment::DELAY)
            continue;

        auto shm = segment.entry.shm();
        // if delayed segment busy, ignore segment, clean up at end...
        if(++shm->count > shm->lines) {
            --shm->count;
            continue;
        }

        auto cid = segment.invite(from, to, subject, [&segment, &sdp](auto msg){
            allow_local(msg);
            if(mode == Mode::ROUTED) {
                auto route = segment.entry.local_to() + ";lr";
                logger.debug(5, "invite record route {}", route);
                osip_message_set_record_route(msg, route.c_str());
            }
            if(!sdp.empty()) {
                osip_message_set_body(msg,sdp.c_str(), sdp.size());
                osip_message_set_content_type(msg, "application/sdp");

            }
        });

        if(cid == -1)
            --shm->count;
        else {
            ++count;
            segment.state = Segment::INVITE;
            sessions.emplace(cid, &segment);
        }
    }

    // remove any delayed segments we failed to invite...
    call.segments.remove_if([&call](const auto& segment) {
        return segment.state == Segment::DELAY && segment.ring_at <= call.rings;
    });
    if(count > 0) {
        logger.debug(3, "call 0x{:08x} delayed invited {} on ring {}", call.shm->id, count, call.rings);
        return true;
    }
    return false;
}

void call_ringer(uint64_t call_id, sync_timepoint ring_timer) {
    // call_id is assumed valid since already verified in exec_calls
    auto& call = calls.at(call_id);
    if(call.shm->state == pbx_call::RINGING) {
        ++call.rings;
        logger.debug(4, "ringing call 0x{:08x} {} times", call_id, call.rings);
        delayed_ringer(call, call.rings);
        ring_timer += std::chrono::seconds(ring);  // synchronized timing...
        ringer.emplace_back(ring_timer, call_id);
    }
}

// call managers

void close_call(const event_t& event) {
    auto cid = event.cid();
    auto did = event.did();
    auto status = event.status();
    segment_t *leg = nullptr;

    if(status == 401 || status == 407)
        status = SIP_REQUEST_TERMINATED;

    try {
        leg = sessions.at(cid);
    }
    catch(...) {
        logger.debug(5, "no call leg for close, cid={}, did={}", cid, did);
        return;
    }

    // if leg closed already, ignore...
    if(leg->state == Segment::CLOSE)
        return;

    auto& call = calls.at(leg->call_id);
    auto code = call.close_status();
    auto call_id = call.shm->id;

drop:
    // close entire call if nothing valid remains
    if(call.segments.size() < 3 || (call.shm->active < 3 && leg->active)) {
        if(call.shm->state == pbx_call::CLOSING)    // if already closed skip
            return;

        cdr_logger(call, status);
        call.shm->active = call.shm->ringing = 0;
        call.shm->state = pbx_call::CLOSING;
        logger.debug(3, "closing call 0x{:08x}, status={}, skip={}", call_id, status, cid);
        for(auto& segment : call.segments) {
            logger.debug(4, "closing segment {}/{}", segment.cid, segment.user_id);
            segment.close(cid == segment.cid ? 0 : code);
        }
        return;
    }

    switch(status) {
    case SIP_BUSY_HERE:
    case SIP_BUSY_EVRYWHERE:
        ++call.busy;
        break;
    case SIP_DECLINE:
    case SIP_REQUEST_TERMINATED:
        ++call.declines;
        break;
    default:
        break;
    }

    auto it = call.segments.begin();
    while(it != call.segments.end()) {
        if(cid == it->cid) {
            auto& segment = *it;
            if(segment.active)
                --call.shm->active;
            else if(segment.state == Segment::RING)
                --call.shm->ringing;
            logger.debug(3, "closing call segment {}/{}", cid, segment.user_id);
            segment.close(0);
            // remove segment since segments size() can determine a call end...
            call.segments.erase(it);
            sessions.erase(cid);
            break;
        }
        ++it;
    }

    // if only originating segment ringing, see if delayed can be advanced
    if(call.shm->ringing == 1 && call.rings < delayed) {
        if(!delayed_ringer(call, delayed))
            goto drop;
    }
}

void drop_call(uint64_t call_id, const reason_t& reason) {
    // In case call has already been closed...
    if(calls.count(call_id) < 1) {
        logger.debug(4, "ignored drop 0x{:08x}", call_id);
        return;
    }

    auto& call = calls.at(call_id);
    if(call.shm->state == pbx_call::CLOSING)
        return;

    logger.debug(3, "dropping call 0x{:08x}, code={}, reason={}", call_id, reason.code, reason.text);
    cdr_logger(call, reason.code);
    call.shm->active = call.shm->ringing = 0;
    call.shm->state = pbx_call::CLOSING;
    for(auto& segment : call.segments) {
        logger.debug(4, "closing call segment {}/{}", segment.cid, segment.user_id);
        segment.close(reason.code, reason.text.empty() ? nullptr : reason.text.c_str());
    }
}

void drop_all() {
    auto reason =reason_t(SIP_REQUEST_TERMINATED, "shutdown");
    for(auto [call_id, call] : calls) {
        for(auto& segment : call.segments) {
            logger.debug(4, "closing call segment {}/{}", segment.cid, segment.user_id);
            segment.close(reason.code, reason.text.empty() ? nullptr : reason.text.c_str());
        }
        logger.debug(3, "dropping call 0x{:08x}, code={}, reason={}", call_id, reason.code, reason.text);
        cdr_logger(call, reason.code);
        call.shm->active = call.shm->ringing = 0;
        call.shm->state = pbx_call::CLOSING;
        call.shm->id = 0;
        freeipc.push_front(call.shm);
    }
    calls.clear();
}

void new_call(const args_t& args) {
    // Make sure we have a place for this call...
    const auto& event = args.event;
    const auto& entry = args.entry;
    const auto& feature = args.feature;
    auto call_id = args.call_id;

    if(freeipc.empty()) {
        event.fail_call(SIP_BUSY_EVRYWHERE);
        return;
    }

    // initialize our objects...
    call_t call(call_id, freeipc.front());
    segment_t segment(call_id, event, entry);
    segment.active = true;
    freeipc.pop_front();

    // calling info...
    auto req = event.request();
    auto caller = "unknown", dialed = "unknown", remote = "";
    if(req && req->from && req->from->url && req->from->url->username)
        caller = req->from->url->username;
    if(req && req->to && req->to->url && req->to->url->username)
        dialed = req->to->url->username;

    str_copy(call.shm->caller, sizeof(call.shm->caller), caller);
    str_copy(call.shm->dialed, sizeof(call.shm->dialed), dialed);
    str_copy(call.shm->remote, sizeof(call.shm->remote), remote);

    // construct call manager states
    if(!is(feature) || feature->name.empty())
        call.to = entry.to(event);
    else
        call.to = format("<{}>", entry.local_alias(feature->name));
    call.from = entry.from(event);
    call.segments.emplace_back(segment);
    calls.emplace(call_id, call);
    sessions.emplace(event.cid(), &calls.at(call_id).segments.back());
    logger.debug(3, "create call 0x{:08x} from {}, capacity={}", call_id, entry.value(), freeipc.size());
    auto body = event.body();
    logger.debug(6, "sdp:\n{}\n", std::string_view(body->body, body->length));
}

void add_segment(const args_t& args) {
    auto call_id = args.call_id;
    auto ring_at = args.rings;
    auto& entry = args.entry;
    auto& call = calls.at(call_id);
    const segment_t segment(call_id, entry, ring_at);

    // Add segment to call...
    call.segments.emplace_back(segment);
    logger.debug(3, "add segment 0x{:08x} to {}", call_id, entry.value());
}

void start_invites(const args_t& args) {
    const auto& feature = args.feature;
    const auto& entry = args.entry;
    const auto& event = args.event;             // originating event...
    auto call_id = args.call_id;
    auto body = event.body();                   // originators sdp
    auto period = event.period();               // expiration period of invite
    auto& call = calls.at(call_id);
    auto from = call.from.c_str();
    auto to = call.to.c_str();
    auto subject = call.subject.c_str();
    auto invited = 0U;

    if(period) {
        call.period = period;
        time(&call.expires);
        call.expires += period;
    }

    if(body->body)
        call.sdp = std::string{content_view(body)};

    // invite newly added segments
    for(auto& segment : call.segments) {
        if(segment.cid != -1 || segment.ring_at > 0 || segment.state == Segment::DELAY)
            continue;

        auto cid = segment.invite(from, to, subject, [&feature, &segment, &body](auto msg){
            allow_local(msg);
            if(is(feature) && !feature->extra.empty())
                osip_message_set_header(msg, "X-Digits", feature->extra.c_str());
            if(is(feature) && !feature->name.empty()) {
                osip_message_set_header(msg, "X-Feature-Script", feature->name.c_str());
                if(feature->sysop)
                    osip_message_set_header(msg, "X-Feature-Mode", "admin");
                else if(is(feature->server))
                    osip_message_set_header(msg, "X-Feature-Mode", "user");
            }

            if(mode == Mode::ROUTED) {
                auto route = segment.entry.local_to() + ";lr";
                logger.debug(5, "invite record route {}", route);
                osip_message_set_record_route(msg, route.c_str());
            }

            if(body->body) {
                osip_message_set_body(msg, body->body, body->length);
                osip_message_set_content_type(msg, "application/sdp");

            }
        });

        if(cid == -1)
            --segment.entry.shm()->count;
        else {
            sessions.emplace(cid, &segment);
            ++invited;
        }
    }

    // clear dead segments...
    call.segments.remove_if([](const auto& segment){return segment.cid == -1 && segment.state != Segment::DELAY;});

    // TODO: Skip and start ringing on first ringing semgnent!
    if(call.segments.size() < 2) {
        event.fail_call(SIP_BUSY_HERE);
        return;
    }

    // if delayed ringing and no invites...
    if(!invited) {
        if(!delayed_ringer(call, delayed)) {
            call.segments.remove_if([](const auto& segment){return !segment.active;});
            ++call.busy;
            event.fail_call(SIP_BUSY_HERE);
        }
    }

    logger.debug(3, "ring invite 0x{:08x} to {}", call_id, entry.value());
}

void reload(const args_t& args) {
    const auto& keys = args.config;
    const auto& config = keys.at("calls");
    const auto& call_mode = config.at("mode");
    display_names = keys.at("display");

    try {
        ring = std::stoi(config.at("ring"));
    }
    catch(...) {
        ring = 4;
    }

    try {
        delayed = std::stoi(config.at("delayed")) / ring;
    }
    catch(...) {
        delayed = 12 / ring;
    }

    if(call_mode == "release")
        mode = Mode::RELEASE;
    else if(call_mode == "route")
        mode = Mode::ROUTED;
}

auto dispatch(const args_t& args, actor_t actor) {
    sync_ptr msgs(msglist);
    if(stopping)
        return;

    msgs->emplace_back(actor, args);
    msgs.unlock();
    cvar.notify_one();
}

} // end namespace

auto server::next_callid() -> uint64_t {
    time_t now{};
    time(&now);

    if(++active_count >= call_entries) {
        --active_count;
        return 0;
    }
    return (static_cast<uint64_t>(++sequence) & 0xfff) | (static_cast<uint64_t>(now) << 12);
}

void server::exec_calls(const keyfile& config) {
    using namespace process;

    auto var = args_t(config);
    auto shm = shm_open(SERVER_IPCPATH ".calls", O_RDWR | O_CREAT, 0660);
    if(shm < 0 || ftruncate(shm, mem_size) < 0)
        logger.crit(97, "failed to allocate calls");
    auto mem = map_t(shm, mem_size, true);
    ::close(shm);
    if(!is(mem))    // NOLINT
        logger.crit(97, "failled to map calls");

    memset(mem.data(), 0, mem_size);
    mapped = static_cast<pbx_call *>(mem.data());
    for(auto idx = size_t(0); idx < call_entries; ++idx)
        freeipc.push_back(&mapped[idx]);

    reload(var);
    logger.info("starting calls thread");
    while(!stopping) {
        sync_ptr msgs(msglist);
        if(!msgs->empty()) {
            auto copy = std::move(msgs->front());
            auto [actor, args] = copy;
            msgs->pop_front();
            msgs.unlock();
            try {
                actor(args);
            }
            catch(const std::exception& e) {
                logger.error("call actor error: {}", e.what());
            }
            continue;
        }

        if(!running)
            ringer.clear();
        else {
            auto now = sync_clock();
            if(!ringer.empty() && ringer.front().first <= now) {
                msgs.unlock();

                auto [ring_timer, call_id] = ringer.front();    // NOTE: refs
                try {
                    if(calls.count(call_id))
                        call_ringer(call_id, ring_timer);
                }
                catch(const std::exception& e) {
                    logger.error("call ringer error: {}", e.what());
                }
                // has to be at end because call_id and ring_timer are references...
                ringer.pop_front();
                continue;
            }
        }

        if(!ringer.empty())
            cvar.wait_until(msgs, ringer.front().first);
        else
            cvar.wait(msgs);
    }

    logger.info("stopping calls thread");
}

void exec::reload_calls(const keyfile& config) {
    dispatch(config, &reload);
}

void exec::hangup_call(uint64_t id, const reason_t& reason) {
    dispatch({id, reason}, [](const args_t& args) {
        drop_call(args.call_id, args.reason);
    });
}

void exec::close_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        close_call(args.event);
    });
}

void exec::stop_calls() {
    dispatch(args_t(), []([[maybe_unused]]const args_t& args) {
        stopping = true;
        drop_all();
    });
}

void exec::call_info(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto type = event.content();
        auto body = event.body();
        auto cid = event.cid();
        auto tid = event.tid();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(4, "no call leg for info, cid={}, tid={}", cid, tid);
            event.fail_call(488);
            return;
        }

        auto& call = calls.at(leg->call_id);
        call.apply([&type, &body](const segment_t& segment){
            send_info(segment.context, segment.did, type.c_str(), body);
            return true;
        }, cid);

        reply_event(event, SIP_OK);
    });
}

void exec::create_call(uint64_t id, const event_t& event, const registry& entry) {
    dispatch({id, event, entry}, &new_call);
}

void exec::invite_call(uint64_t id, const registry& entry, unsigned rings) {
    dispatch({id, entry, rings}, &add_segment);
}

void exec::invite_delayed(uint64_t id, const registry& entry, unsigned rings) {
    dispatch({id, entry, rings}, [](const args_t& args) {
        auto call_id = args.call_id;
        auto ring_at = args.rings;
        auto& entry = args.entry;
        auto& call = calls.at(call_id);

        if(!ring_at)
            ring_at = delayed;

        segment_t segment(call_id, entry, ring_at);
        segment.state = Segment::DELAY;

        // Add segment to call...
        call.segments.emplace_back(segment);
        logger.debug(3, "add segment 0x{:08x} to {}", call_id, entry.value());
    });
}

void exec::start_call(uint64_t id, const event_t& event, const registry& entry) {
    dispatch({id, event, entry}, &start_invites);
}

void exec::call_server(uint64_t id, const event_t& event, const feature_t& feature) {
    dispatch({id, event, feature}, &new_call);
    dispatch({id, feature->server, 0}, &add_segment);
    dispatch({id, event, feature}, &start_invites);
}

void exec::release_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto cid = event.cid();
        auto did = event.did();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(5, "no call leg for release, cid={}, did={}", cid, did);
            return;
        }

        if(leg->state != Segment::CLOSE) // should have already been closed...
            close_call(event);

        auto& call = calls.at(leg->call_id);

        // if not closing full call, find segment to release...
        if(call.shm->state != pbx_call::CLOSING) {
            logger.debug(3, "releasing segment {}/{} from call", cid, leg->user_id);
            call.segments.remove_if([&cid](const auto& segment){return segment.cid == cid;});
            sessions.erase(cid);
        }
        else {
            auto call_id = call.shm->id;
            call.shm->id = 0;
            freeipc.push_front(call.shm);
            for(auto& segment : call.segments) {
                logger.debug(4, "releasing call segment {}/{}", segment.cid, segment.user_id);
                sessions.erase(segment.cid);
            }
            calls.erase(call_id);
            logger.debug(3, "releasing call 0x{:08x}", call_id);
        }
    });
}

void exec::trying_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto cid = event.cid();
        auto tid = event.tid();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(5, "no call leg for ringing, cid={}, tid={}", cid, tid);
            return;
        }

        if(leg->state != Segment::INVITE && leg->state != Segment::RING)
            return;

        auto call_id = leg->call_id;
        if(tid > 0)
            leg->tid = tid;
        logger.debug(3, "trying call 0x{:08x}, cid={}, tid={}", call_id, cid, tid);
    });
}

void exec::failed_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto cid = event.cid();
        auto did = event.did();
        auto status = event.status();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(5, "no call leg for failing, cid={}, did={}", cid, did);
            return;
        }

        logger.debug(4, "failed call segment {}/{}, status={}", cid, leg->user_id, status);
        auto call_id = leg->call_id;
        auto& call = calls.at(call_id);
        switch(status) {
        case SIP_DECLINE:
            ++call.declines;
            break;
        case SIP_BUSY_EVRYWHERE:
        case SIP_BUSY_HERE:
            ++call.busy;
        default:
            break;
        }

        close_call(event);
        if(call.shm->state != pbx_call::CLOSING) {
            logger.debug(5, "removing failed segment {}/{} from call 0x{:08x}", cid, leg->user_id, call_id);
            call.segments.remove_if([&cid](const auto& segment){return segment.cid == cid;});
            sessions.erase(cid);
        }
        else
            logger.debug(5, "removing failed call 0x{:08x}", call_id);
    });
}

void exec::ack_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto cid = event.cid();
        auto did = event.did();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(5, "no call leg for ack, cid={}, did={}", cid, did);
            return;
        }

        // if already aknowledged, skip
        if(leg->ackflag)
            return;

        auto call_id = leg->call_id;
        auto& call = calls.at(call_id);
        leg->ackflag = true;
        if(did > 0) {
            leg->did = did;
            leg->tid = -1;
        }

        logger.debug(3, "ack call 0x{:08x}, cid={}, did={}", call_id, cid, did);
        for(auto& segment : call.segments) {
            // if not active, or already acked, no need...
            if(!segment.active || segment.ackflag)
                continue;

            if(!segment.ack())
                logger.warn("ack segment {}/{} failed; tid={}", segment.cid, segment.user_id, segment.tid);
            else
                logger.debug(4, "ack sent to segment {}/{}, did={}, tid={}", segment.cid, segment.user_id, segment.did, segment.tid);
            segment.tid = -1;
        }
    });
}

void exec::answer_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto cid = event.cid();
        auto did = event.did();
        auto body = event.reply();
        auto period = event.refresh();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(5, "no call leg for answer, cid={}, did={}", cid, did);
            return;
        }

        logger.debug(4, "answer call segment {}/{}, did={}", cid, leg->user_id, did);
        if(leg->active && leg->reinvited > 0)
            return;

        auto call_id = leg->call_id;
        auto& call = calls.at(call_id);

        assert(call.shm->active > 0);
        assert(call.segments.size() > 1);

        // update call periods as needed based on answer...
        if(period) {
            call.period = period;
            time(&call.expires);
            call.expires += period;
        } else
            call.period = call.expires = call.minimal = 0;

        // process reinvite to originating party and return...
        if(leg->reinvited) {
            call.apply([&body](segment_t& segment){
                logger.debug(5, "reinvite originating segment {}/{}",
                    segment.cid, segment.user_id);
                auto result = segment.answer(SIP_OK, [&body](auto msg){
                    if(body && body->body) {
                        osip_message_set_body(msg, body->body, body->length);
                        osip_message_set_content_type(msg, "application/sdp");
                    } else print("*** NO SDP\n");
                });
                segment.tid = -1;
                return result;
            }, cid);
            leg->reinvited = false;
            return;
        }

        // process new call answer...
        ++call.shm->active;
        leg->active = true;
        leg->did = did;
        if(leg->state == Segment::RING)
            --call.shm->ringing;
        leg->state = Segment::JOINED;

        for(auto& segment : call.segments) {
            // create answer for pending extensions...
            if(segment.active && segment.state == Segment::RING) {
                logger.debug(5, "answer pending segment {}/{}",
                    segment.cid, segment.user_id);
                segment.answer(SIP_OK, [&body](auto msg){
                    if(body && body->body) {
                        logger.debug(6, "sdp:\n{}\n", std::string_view(
                            body->body, body->length));
                        osip_message_set_body(msg, body->body, body->length);
                        osip_message_set_content_type(msg, "application/sdp");
                    }
                });
                --call.shm->ringing;
                segment.state = Segment::JOINED;
            }
            // skip active segments
            if(segment.active)
                continue;
            // shutdown other pending segments
            if(segment.state == Segment::RING)
                --call.shm->ringing;
            sessions.erase(segment.cid);
            segment.close(SIP_REQUEST_TERMINATED);
        }
        // connect and kill pending segments completely...
        time(&call.joined);
        call.shm->state = pbx_call::CONNECTED;
        call.segments.remove_if([](const auto& segment) {return !segment.active;});
        assert(call.segments.size() > 1);
    });
}

void exec::ringing_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto cid = event.cid();
        auto did = event.did();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(5, "no call leg for ringing, cid={}, did={}", cid, did);
            return;
        }

        if(leg->state != Segment::INVITE)
            return;

        auto call_id = leg->call_id;
        leg->did = did;
        leg->state = Segment::RING;
        auto& call = calls.at(call_id);
        if(call.shm->state == pbx_call::INVITING) {
            call.shm->state = pbx_call::RINGING;
            auto count = call.apply([](const segment_t& segment){
                return segment.answer(SIP_RINGING);
            });
            create_ringer(call_id);
            logger.debug(3, "ringback call 0x{:08x} to {} segment(s)", call_id, count);
        }
        ++call.shm->ringing;
    });
}

void exec::redo_invite(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto context = event.context();
        auto type = event.content();
        auto cid = event.cid();
        auto did = event.did();
        auto tid = event.tid();
        auto body = event.body();
        auto period = event.refresh();
        segment_t *leg = nullptr;

        try {
            leg = sessions.at(cid);
        }
        catch(...) {
            logger.debug(5, "no call leg for reinvite, cid={}, did={}", cid, did);
            event.fail_call(488);
            return;
        }

        logger.debug(4, "reinvite call segment {}/{}, did={}", cid, leg->user_id, did);
        auto call_id = leg->call_id;
        auto& call = calls.at(call_id);

        if(!leg->active || call.shm->active < 2) {  // has to be an active call...
            osip_message_t *msg = nullptr;
            const exosip::guard_t lock(context);
            eXosip_call_build_answer(context, tid, 500, &msg);
            if(leg->active && msg)
                osip_message_set_header(msg, "Reply-After", "8");
            eXosip_call_send_answer(context, tid, 500, msg);
            return;
        }

        if(period) {
            call.period = period;
            time(&call.expires);
            call.expires += period;
        } else
            call.period = call.expires = call.minimal = 0;

        call.apply([&type, &body](segment_t& segment){
            send_invite(segment.context, segment.did, type.c_str(), body);
            segment.reinvited = true;
            return true;
        }, cid);

        // waits for answer from re-inviting segment
        leg->tid = tid;
    });
}

void exec::hangup_user(const registry& entry, const reason_t& reason) {
    dispatch({entry, reason}, [](const args_t& args) {
        const auto& user = args.entry;
        const auto& reason = args.reason;
        logger.debug(5, "dropping user {}, code={}, reason={}", user.value(), reason.code, reason.text);

        for(auto& [id, call] : calls) {
            if(call.shm->pbx_call::state == pbx_call::CLOSING)
                continue;

            for(auto& segment : call.segments) {
                if(segment.entry.extension() != user.extension())
                    continue;

                if(segment.state == Segment::CLOSE)
                    break;

                if(call.segments.size() < 3 || (call.shm->active < 3 && segment.active)) {
                    drop_call(id, reason);
                    break;
                }
                logger.debug(4, "dropping user {} from call 0x{:08x}", user.value(), id);
                if(segment.active)
                    --call.shm->active;
                else if(segment.state == Segment::RING)
                    --call.shm->ringing;
                segment.close(reason.code, reason.text.empty() ? nullptr : reason.text.c_str());
                break;
            }
        }
    });
}

