/*
 * Copyright (C) 2023 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"
#include "features.hpp"
#include "calls.hpp"

#include <moderncli/keyfile.hpp>

#include <algorithm>
#include <csignal>
#include <list>

using namespace server;

namespace {
struct args_t final {
    args_t() = default;
    args_t(keyfile keys) : config(std::move(keys)) {}
    args_t(feature_t f, event_t e) : feature(std::move(f)), event(std::move(e)) {}

    feature_t feature;
    event_t event;
    keyfile config;
};

using actor_t = void (*)(const args_t&);
using msg_t = std::pair<actor_t, args_t>;

bool stopping = false;
std::condition_variable cvar;
unique_sync<std::list<msg_t>> msglist;

void feature_echo(const args_t& args) {
    auto& event = args.event;
    auto& feature = args.feature;
    auto& digits = feature->extra;
    auto& user = feature->origin;
    auto from = std::to_string(user.extension());

    if(!digits.empty())
        system_message("Echo digits = " + digits, from);
    logger.debug(2, "echo request from {}, digits=\"{}\"", from, digits);
    event.terminate_call();
}

void feature_reload(const args_t& args) {
    auto& event = args.event;
    event.terminate_call();
    raise(SIGHUP);
}

void feature_dialer(const args_t& args) {
    auto& event = args.event;
    auto& feature = args.feature;
    auto& user = feature->origin;
    auto& invite = feature->server;
    auto up = user.shm();
    auto display = user.contact().display();

    // make sure we have something to invite...
    if(!is(invite) || !invite.allow_invite()) {
        event.fail_call(SIP_NOT_FOUND);
        return;
    }

    // Make sure we have capacity...
    if(++up->count > up->lines) {
        --up->count;
        event.fail_call(SIP_BUSY_HERE);
        return;
    }


    // check for busy...
    auto ep = invite.shm();
    if(++ep->count >  ep->lines) {
        --ep->count;
        event.fail_call(SIP_BUSY_HERE);
        return;
    }
    logger.debug(4, "call from \"{}\" for feature {}", display, feature->name);

    auto call_id = next_callid();
    if(!call_id) {
        --up->count;
        event.fail_call(SIP_BUSY_EVRYWHERE);
    }

    exec::call_server(call_id, event, feature);
}

// TODO: see if any of these can be reduced to a string view
std::unordered_map<std::string, actor_t> features;
std::unordered_map<std::string, std::string> used;
std::unordered_map<std::string, actor_t> names = {
    {"echo", &feature_echo},
    {"reload", &feature_reload},
};

void reload(const args_t& args) {
    const auto& keys = args.config;
    auto codes = keys.at("features");

    features.clear();
    used.clear();
    std::for_each(codes.begin(), codes.end(), [](const auto& pair) {
        const auto& [code, name] = pair;
        if(!name.empty() && name[0] == '@') {
            features[code] = &feature_dialer;
            used[code] = name.c_str() + 1;
        }
        else {
            auto exec = names[name];
            if(exec == nullptr)
                return;
            features[code] = exec;
            used[code] = name;
        }
    });
    logger.debug(4, "loaded {} feature handler(s)", features.size());
}

void dispatch(args_t args, actor_t actor) {
    sync_ptr msgs(msglist);
    if(stopping)
        return;

    msgs->emplace_back(actor, args);
    msgs.unlock();
    cvar.notify_one();
}
} // end anon namespace

void server::exec_features(const keyfile& config) {
    auto args = args_t(config);

    reload(args);
    logger.info("starting features thread");
    while(!stopping) {
        sync_ptr msgs(msglist);
        if(!msgs->empty()) {
            auto copy = std::move(msgs->front());
            auto [actor, parms] = copy;
            msgs->pop_front();
            msgs.unlock();
            try {
                actor(parms);
            }
            catch(const std::exception& e) {
                logger.error("features actor error: {}", e.what());
            }
            continue;
        }
        cvar.wait(msgs);
    }
    logger.info("stopping features thread");
}

void exec::reload_features(const keyfile& config) {
    dispatch(config, &reload);
}

void exec::stop_features() {
    dispatch(args_t(), []([[maybe_unused]]const args_t& args){
        stopping = true;
    });
}

void exec::dialed_feature(feature_t feature, const event_t& event) {
    auto dialed = feature->code;
    for(const auto& pair : features) {
        const auto& [dial, exec] = pair;
        auto len = dial.length();
        if(dialed.length() >= len && dialed.substr(0, len) == dial) {
            feature->name = used[dial];
            if (dialed.length() > len)
                feature->extra = dialed.substr(len);
            dispatch({feature, event}, exec);
            return;
        }
    };
    event.fail_call(SIP_NOT_FOUND);
}
