/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "nonce.hpp"

#include <moderncli/strings.hpp>
#include <moderncli/random.hpp>

using namespace server;
using seconds_t = std::chrono::seconds;

namespace {
auto offset = static_cast<uint32_t>(system_clock() & 0x0fffff);
auto started = sync_clock();
std::atomic<uint32_t> counter;  // each nonce unique
}

nonce_t::nonce_t() {    // NOLINT
    data_.timestamp = ntohl(current_timestamp());
    data_.sequence = ntohl(counter.fetch_add(1));
    crypto::rand(data_.random_iv, sizeof(data_.random_iv));
    key_ = to_hex(data(), 16);
    created_ = sync_clock();
}

nonce_t::nonce_t(const std::string_view& from) { // NOLINT
    if(!from_hex(from, &data_.buffer[0], 16))
        memset(&data_.buffer[0], 0, 16);
    else
        key_ = from.substr(0, 32);
    created_ = sync_clock();
}

auto nonce_t::operator=(const std::string_view& from) -> nonce_t& {
    if(!from_hex(from, &data_.buffer[0], 16))
        memset(&data_.buffer[0], 0, 16);
    else
        key_ = from.substr(0, 32);
    created_ = sync_clock();
    return *this;
}

auto nonce_t::hash() const -> size_t {
    size_t result = 0;
    auto count = 16U;
    auto dp = &data_.buffer[0];
    while(count--) {
        result <<= 2;
        result ^= *dp;
        ++dp;
    }
    return result;
}

auto nonce_t::current_timestamp() -> uint32_t {
    auto elapsed = sync_clock() - started;
    return static_cast<uint32_t>(std::chrono::duration_cast<seconds_t>(elapsed).count()) + offset;
}
