/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"
#include "registry.hpp"
#include "calls.hpp"
#include "messages.hpp"
#include "features.hpp"
#include "network.hpp"

#include <moderncli/keyfile.hpp>
#include <moderncli/process.hpp>
#include <moderncli/random.hpp>

#include <unordered_set>
#include <csignal>
#include <list>

using namespace std::literals::string_literals;
using namespace server;

namespace {
inline constexpr auto reg_size = sizeof(pbx_reg) * registry_entries;
inline constexpr auto mem_size = reg_size + sizeof(pbx_sys);
inline constexpr auto help_text = "*** system commands: /echo /msg /who";

struct args_t final {
    args_t() = default;
    args_t(keyfile cfg) : config(std::move(cfg)) {}
    args_t(event_t evt) : event(std::move(evt)) {}
    args_t(pbx_var evt) : ipc(std::move(evt)) {}

    keyfile config;
    event_t event;
    pbx_var ipc;
};

using actor_t = void (*)(const args_t&);
using msg_t = std::pair<actor_t, args_t>;

std::atomic<int> active{0};
std::unordered_map<std::string, registry> clients;
std::array<registry, registry_entries> mapped;
registry empty_registry;
std::string hostname = "localhost";
std::string realm = "Coventry";
std::string algorithm = "SHA-256,MD5";
std::string state = "none";
std::string sitename = "Coventry Server";
std::string sitefrom = "sip:system@locahost:5060";
uint16_t hostport = 5060;
volatile bool timer = false;
pbx_reg *reg_shared = nullptr;
pbx_sys *reg_system = nullptr;
keyfile::keys server_messages, display_names;
time_t max_expires = 300;
keyfile key_storage;
std::thread::id thread_id;
auto key_updates = true;    // always save after initial load
auto stopping = false;
std::string feature_server, assist_server;
std::string webserver;

bool initial = true;
std::condition_variable cvar;
unique_sync<std::list<msg_t>> msglist;

void dispatch(const args_t& args, actor_t actor) {
    sync_ptr msgs(msglist);
    if(stopping)
        return;

    msgs->emplace_back(actor, args);
    msgs.unlock();
    cvar.notify_one();
}


class registry_map final {
public:
    registry_map() { // NOLINT
        memset(data_, 0, sizeof(data_));
    }

    explicit registry_map(std::string_view list) { // NOLINT
        memset(data_, 0, sizeof(data_));
        add(list);
    }

    auto operator[](unsigned index) const -> bool {
        if(index < dialplan_min || index > dialplan_max)
            throw std::range_error("outside extension map");

        index -= dialplan_min;
        const uint8_t mask = 0x01 << index % 8;
        return data_[index >> 3] & mask;
    }

    auto operator[](const registry_map& map) const {
        for(auto index = 0UL; index < sizeof(data_); ++index) {
            if((data_[index] & map.data_[index]) != map.data_[index])
                return false;
        }
        return true;
    }

    auto operator+=(unsigned value) -> registry_map& {
        operator()(value, true);
        return *this;
    }

    auto operator-=(unsigned value) -> registry_map& {
        operator()(value, false);
        return *this;
    }

    auto operator==(unsigned value) const {
        return operator[](value);
    }

    auto operator!=(unsigned value) const {
        return !operator[](value);
    }

    auto operator()(unsigned ext) const {
        return operator[](ext);
    }

    auto data() const {
        return &data_;
    }

    auto operator&=(const registry_map& from) {
        for(auto index = 0UL; index < sizeof(data_); ++index)
            data_[index] &= from.data_[index];
        return *this;
    }

    auto operator|=(const registry_map& from) {
        for(auto index = 0UL; index < sizeof(data_); ++index)
            data_[index] |= from.data_[index];
        return *this;
    }

    auto operator^=(const registry_map& from) {
        for(auto index = 0UL; index < sizeof(data_); ++index)
            data_[index] ^= from.data_[index];

        return *this;
    }

    void operator()(unsigned index, bool state) {
        if(index < dialplan_min || index > dialplan_max)
            throw std::range_error("outside extension map");

        index -= dialplan_min;
        const uint8_t mask = 0x01 << index % 8;
        if(state)
            data_[index >> 3] |= mask;
        else
            data_[index >> 3] &= ~mask;
    }

    auto is_sync(unsigned index) const {
        if(index < dialplan_min || index > dialplan_max)
            return false;

        auto offset = index - dialplan_min;
        const uint8_t mask = 0x01 << (offset % 8);

        offset >>= 3;
        auto byte = data_[offset];
        while(byte != data_[offset])
            byte = data_[offset];

        return (byte & mask) != 0;
    }

    auto count() const {
        auto count = 0U;
        auto index = dialplan_min;
        while(index <= dialplan_max) {
            auto offset = index - dialplan_min;
            auto mask = 0x01U;
            auto byte = data_[offset / 8];
            if(!byte) { // skip empty optimization...
                index += 8;
                continue;
            }
            for(auto bits = 0; bits < 8; ++bits) {
                if(byte & mask)
                    ++count;

                ++index;
                mask <<= 1;
            }
        }
        return count;
    }

    auto apply(std::function<bool(registry&)> pred) const {
        auto total = 0U;
        auto index = dialplan_min;
        while(index <= dialplan_max) {
            auto offset = index - dialplan_min;
            auto mask = 0x01U;
            auto byte = data_[offset / 8];
            if(!byte) { // skip empty optimization...
                index += 8;
                continue;
            }
            for(auto bits = 0; bits < 8; ++bits) {
                if(byte & mask) {
                    auto& entry = mapped.at(offset + bits);
                    if(is(entry) && pred(entry))
                        ++total;
                }
                ++index;
                mask <<= 1;
            }
        }
        return total;
    }

    auto empty() const {
        return std::all_of(data_, data_ + sizeof(data_), [](auto bits) {
            return !bits;
        });
    }

    operator bool() const {
        return !empty();
    }

    auto operator!() const {
        return empty();
    }

    auto clear() {
        memset(data_, 0, sizeof(data_));
    }

    void add(std::string_view list) {
        auto tokens = tokenize(list, ",:; ");
        for(const auto& token : tokens) {
            auto value = util::get_value(token) - dialplan_min;
            if(value < 0 || value >= registry_entries)
                continue;
            auto mask = 0x01 << (value % 8);
            value /= 8;
            data_[value] |= mask;
        }
    }

    static auto size() {
        return sizeof(data_);
    }

private:
    uint8_t data_[registry_entries / 8];
};

auto operator&(const registry_map& l, const registry_map& r) {
    registry_map tmp(l);
    tmp &= r;
    return tmp;
}

[[maybe_unused]] auto operator|(const registry_map& l, const registry_map& r) {
    registry_map tmp(l);
    tmp |= r;
    return tmp;
}

auto operator^(const registry_map& l, const registry_map& r) {
    registry_map tmp(l);
    tmp ^= r;
    return tmp;
}

// registry maps we use...

registry_map active_map;
registry_map coverage_map[registry_entries];
registry_map delayed_map[registry_entries];
std::unordered_map<std::string, registry_map> group_map;
std::unordered_map<std::string, registry_map> access_map;
std::unordered_map<std::string, registry_map> room_map;
std::unordered_map<std::string, std::vector<std::string>> zone_map;

// subscriptions

class subscriptions;

struct subscription_t final {
    struct subscriber {
        time_t expires{0};
        int dialog{-1};
    };

    std::unordered_set<unsigned> subscribed;
    std::array<subscriber, registry_entries> subscribers;
    subscriptions* parent{nullptr};
    std::string body;
    unsigned count{0};
    time_t expires{0};    // published body expires at...

    auto apply(std::function<bool(exosip::msg_t msg)> pred) const -> unsigned {
        auto total = 0U;
        auto now = system_clock();
        for(auto pos = 0U; pos < registry_entries; ++pos) {
            const auto& sub = subscribers[pos];
            if(sub.expires <= now) // if dead or expired...
                continue;

            const auto& entry = mapped.at(pos);
            if(!is(entry))
                continue;

            auto ctx = entry.context();
            const exosip::guard_t lock(ctx);
            exosip::msg_t msg{nullptr};
            if(msg && pred(msg)) {
                ++total;
            }
        }
        return total;
    }

    void clear() {
        subscribed.clear();
        subscribers.fill(subscriber());
        count = 0;
    }

    auto clear(unsigned ext) -> bool {
        auto offset = ext - dialplan_min;
        auto& sub = subscribers.at(offset);
        if(sub.expires == 0)
            return false;
        sub.expires = 0;
        --count;
        return true;
    }
};

class subscriptions final {
public:
    subscriptions(std::string_view type, std::string_view event) :
    type_(type), event_(event) {}

    auto type() {
        return type_;
    }

    auto event() {
        return event_;
    }

    auto get(unsigned ext) -> subscription_t& {
        if(ext < dialplan_min || ext > dialplan_max)
            throw std::range_error("outside extension map");

        auto offset = ext - dialplan_min;
        return subscriptions_[offset];
    }

    auto release(unsigned ext) {
        if(ext < dialplan_min || ext > dialplan_max)
            throw std::range_error("outside extension map");

        auto offset = ext - dialplan_min;
        auto &entry = subscriptions_[offset];
        auto subs = entry.subscribed;
        auto result = std::all_of(subs.cbegin(), subs.cend(), [&](auto sub){
            auto &member = subscriptions_[sub - dialplan_min];
            return member.clear(ext);
        });
        entry.clear();
        return result;
    }

private:
    std::array<subscription_t, registry_entries> subscriptions_;
    std::string_view type_;
    std::string_view event_;
};

subscriptions presence("application/pidf_xml", "presense");

void publish_event(subscriptions& group, const registry& entry, const event_t& event) {
    if(entry.empty())
        return;

    auto& pubsub = group.get(entry.extension());
    pubsub.parent = &group;
    pubsub.body = std::string{content_view(event.body())};
    pubsub.expires = system_clock(event.period());
}

// helper functions...

void sync_storage() {
    if(!key_updates)
        return;
    key_storage.write("registry.keys");
    key_updates = false;
}

auto system_from() {
    return format("\"{}\" <{}>", sitename, sitefrom);
}

auto is_authorized(registry& entry, const event_t& event) {
    auto keys = entry.keys();
    if(!entry.verify(event))
        return false;

    try {
        if(event.algorithm() == "SHA-256")
            return util::sha256_digest(keys.at("sha256"), event) == event.digest();
        if(event.algorithm() == "MD5")
            return util::md5_digest(keys.at("md5"), event) == event.digest();
    }
    catch(...) {}   // NOLINT

    logger.warn("User {} has no {} secret", event.user(), event.algorithm());
    return false;
}

auto chk_secret(std::string_view user, std::string_view secret, size_t size) {
    if(size != secret.size())
        return false;

    if(std::any_of(secret.cbegin(), secret.cend(), [](char ch){
        return ch < '0' || ch > 'f' || (ch > '9' && ch < 'a');
    })) {
        logger.warn("user {} invalid secret {}", user, secret);
        return false;
    }
    return true;
}

void notify() {
    timer = true;
    cvar.notify_one();
}

auto find(std::string_view user) noexcept -> registry& {
    if(user.empty())
        return empty_registry;

    auto value = 0U;
    for(auto ch : user) {
        if(!isdigit(ch))
            return empty_registry;
        value = (value * 10) + (ch - '0');
    }

    if(value < dialplan_min || value > dialplan_max)
        return empty_registry;

    return mapped.at(value - dialplan_min);
}

[[maybe_unused]] auto find(const event_t& event) noexcept -> registry& { //TODO: use this?
    if(event.authorize())
        return find(event.user());

    const contact_t route(event);
    try {
        return clients.at(route.addr());
    }
    catch(...) {
        return empty_registry;
    }
}

void changed(const registry_map& current) {
    auto prior = active_map;
    active_map = current;

    auto changes = current ^ prior;
    auto dropped = changes & prior;
    //auto added = changes & current;

    for(auto base = 0UL; base < mapped.size(); ++base) {
        auto ext = dialplan_min + base;
        if(!dropped[ext])
            continue;
        logger.debug(4, "registry: drop {}", ext);
        mapped.at(base).drop();
    }
}

auto rescan() {   // fast sync compute...
    registry_map current;
    time_t now{}, latest = 0;
    time(&now);

    logger.debug(2, "registry timeout scan");
    sync_storage();
    for(auto& entry : mapped) {
        auto expires = entry.expires();
        if(!expires)            // if already expired, ignore
            continue;
        if(expires <= now) {     // drop expired entry...
            auto& shared = reg_shared[entry.extension() - dialplan_min];
            shared.activated = shared.expires = 0;

            logger.notice("expiring {}", entry.origin_uri());
            str_copy(shared.agent, sizeof(shared.agent), "-");
            str_copy(shared.name, sizeof(shared.name), "");
            str_copy(shared.id, sizeof(shared.id), "-");
            entry.drop();
            continue;
        }
        current += entry.extension();
        if(expires < latest)
            latest = expires;
    }

    // locked registry
    changed(current);
    return latest;
}

// cli message parsers...

void cli_sysop(const registry& user, std::string_view text) {
    text.remove_prefix(6);
    auto tokens = tokenize(text);
    if(tokens.size() == 1 && tokens[0] == "reload") {
        raise(SIGHUP);
    }
    else {
        exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", "*** unknown sysop command", "text/plain", user.context()));
    }
}

auto cli_command(const registry& user, std::string_view text) {
    auto tokens = tokenize(text);
    if(tokens[0] == "/echo") {
        text.remove_prefix(5);
        while(!text.empty() && text.front() == ' ')
            text.remove_prefix(1);
        exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", text, "text/plain", user.context()));
        return true;
    }

    if(tokens[0] == "/who") {
        auto id = std::to_string(user.extension());
        std::string result, status;
        if(tokens.size() == 1)
            tokens[0] = id;
        else
            tokens.erase(tokens.begin());
        for(auto token : tokens) {
            if(registry::is_extension(token)) {
                auto& entry = find(token);
                if(is(entry)) {
                    status = "here";
                    switch(entry.presence()) {
                    case pbx_reg::GONE:
                        status = "gone";
                        break;
                    case pbx_reg::DND:
                        status = "dnd";
                        break;
                    case pbx_reg::AWAY:
                        status = "away";
                        break;
                    case pbx_reg::HERE:
                        if(entry.shm()->count)
                            status = "busy";
                        break;
                    }
                }
                else
                    status = "gone";
                result = format("Extension {} is {}, status={}, agent={}", entry.value(), entry.contact().display(), status, entry.agent());
            }
            else
                result = format("*** Unknown {}", token);
            exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", result, "text/plain", user.context()));
        }
        return true;
    }
    return false;
}

auto cli_help(const registry& user) {
    exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", help_text, "text/plain", user.context()));
    return true;
}

void reload(const args_t& args) {
    const auto& keys = args.config;
    auto common = keys.at("common");
    auto server = keys.at("server");
    auto groups = keys.at("groups");
    auto access = keys.at("access");
    auto rooms = keys.at("rooms");
    auto zones = keys.at("zones");

    display_names = keys.at("display");
    server_messages = keys.at("messages");
    realm = server["realm"];
    algorithm = server["algorithm"];
    hostname = server["hostname"];
    hostport = stoi(server["port"]);
    sitefrom = "sip:system@" + hostname + ":" + std::to_string(hostport);
    sitename = server["sitename"];
    webserver = format("{}://{}:{}", server["urlschema"], hostname, server["webserver"]);
    state = server["state"];
    max_expires = util::get_expires(server["expires"]);
    uint16_t base = 0;
    time_t now{};

    ++reg_system->series;
    sync_storage();
    registry_map current;
    time(&now);
    logger.debug(4, "registry: reload for realm {} state {}", realm, state);
    str_copy(reg_system->state, sizeof(reg_system->state), state);
    str_copy(reg_system->realm, sizeof(reg_system->realm), realm);
    str_copy(reg_system->digest, sizeof(reg_system->digest), algorithm);

    // clear coverage maps...
    for(auto pos = 0U; pos < registry_entries; ++pos) {
        coverage_map[pos].clear();
        delayed_map[pos].clear();
    }

    // clear and load our named groups...
    group_map.clear();
    for(const auto& [key, value] : groups)
        group_map.emplace(key, value);
    group_map.try_emplace("system", "");    // add system group if missing...
    group_map.try_emplace("sysop", "");  // add sysop group if missing...

    // clear and load access control members...
    access_map.clear();
    for(const auto& [key, value] : access)
        access_map.emplace(key, value);

    // clear and load room list...
    room_map.clear();
    for(const auto& [key, value] : rooms) {
        logger.debug(4, "room '{}' includes {}", key, value);
        room_map.emplace(key, value);
    }

    // clear and load zone list...
    zone_map.clear();
    for(const auto& [key, value] : zones)
        zone_map.emplace(key, tokenize(value, ",:; "));

    // load and activate extension entries
    for(auto& entry : mapped) {
        auto pos = base;
        auto ext = dialplan_min + (base++);
        auto key = std::to_string(ext);
        auto prior = !entry.empty();
        auto exists = keys.exists(key);

        // prune dead keys
        if(!exists && key_storage.exists(key)) {
            key_storage.remove(key);
            key_updates = true;
        }

        if(entry.empty() && !exists)
            continue;

        if(entry.empty())
            entry = registry(ext);

        if(keys.exists(key)) {
            if(prior)
                logger.debug(5, "registry: update {}", ext);
            else if(initial)
                logger.debug(5, "registry: create {}", ext);
            else
                logger.debug(3, "registry: create {}", ext);

            auto pairs = keys.at(key);
            pairs.insert(common.begin(), common.end());
            try {
                auto stored = key_storage.at(key);
                for(const auto& [id, value] : stored)
                    pairs.insert_or_assign(id, value);
            }
            catch(...) {}   // NOLINT

            entry.keys(pairs);
            try {
                auto coverage = pairs.at("coverage");
                coverage_map[pos].add(coverage);
            }
            catch(...) {
                coverage_map[pos] += entry.extension();
            }

            try {
                auto delayed = pairs.at("delayed");
                delayed_map[pos].add(delayed);
            }
            catch(...) {}   // NOLINT
        }
        else {
            if(prior)
                logger.debug(3, "registry: remove {}", ext);

            entry = empty_registry;
        }

        if(entry.expires() > now)
            current += ext;
    }
    changed(current);
}
} // end anon namespace

registry::registry(unsigned key) :
ptr_(std::make_shared<data_t>()), shm_(&reg_shared[key - dialplan_min]) {
    ptr_->extension = key;
    ptr_->count = 0;
    ptr_->contact.user_ = std::to_string(key);
    ptr_->contact.host_ = hostname;
    ptr_->contact.port_ = hostport;
    ptr_->contact.display_ = "Ext. " + ptr_->contact.user_;
    ptr_->welcome = true;

    shm_->count = 0;
    shm_->lines = 0;
    shm_->flags.invite = shm_->flags.message = false;
}

auto registry::is_active() -> bool {
    return active.load() > 0;
}

auto registry::is_extension(std::string_view text) -> bool {
    if(text.empty())
        return false;

    auto ext = 0U;
    for(auto ch : text) {
        if(!isdigit(ch))
            return false;
        ext = (ext * 10) + (ch - '0');
    }

    return ext >= dialplan_min && ext <= dialplan_max;
}

auto registry::local_uri() const -> std::string {
    return ptr_ && context_ ? format("{}:{}", ptr_->schema, contact().id()) : "";
}

auto registry::local_alias(std::string_view to) const -> std::string {
    return ptr_ && context_ ? format("{}:{}@{}", ptr_->schema, to, contact().addr()) : "";
}

auto registry::local_to() const -> std::string {
    return ptr_ && context_ ? format("{}:{}", ptr_->schema, contact().addr()) : "";
}

auto registry::origin_uri() const -> std::string {
    return ptr_ && context_ ? format("{}:{}", ptr_->schema, origin().id()) : "";
}

auto registry::route_to() const -> std::string {
    return ptr_ && context_ ? format("{}:{}", ptr_->schema, origin().addr()) : "";
}

void registry::keys(const keyfile::keys& pairs) {
    if(ptr_) {
        std::string sha256, md5, secret;
        auto user = std::to_string(extension());

        try {
            shm_->lines = std::stoi(pairs.at("lines"));
        }
        catch(...) {
            shm_->lines = 1;
        }

        try {
            secret = pairs.at("secret");
        }
        catch(...) {
            try {
                secret = pairs.at("password");
            }
            catch(...) {}   // NOLINT
        }

        if(!secret.empty()) {
            sha256 = util::sha256_secret(user, realm, secret);
            md5 = util::md5_secret(user, realm, secret);
        }

        const std::lock_guard lock(ptr_->mutex);
        ptr_->keys = pairs;
        if(chk_secret(user, sha256, 64))
            ptr_->keys["sha256"] = sha256;
        if(chk_secret(user, md5, 32))
            ptr_->keys["md5"] = md5;

        try {
            ptr_->caller = pairs.at("caller");
        }
        catch(...) {
            try {
                ptr_->caller = pairs.at("name");
            }
            catch(...) {
                ptr_->caller = "";
            }
        }

        try {
            ptr_->type = pairs.at("type");
        }
        catch(...) {
            ptr_->type = "generic";
        }

        try {
            ptr_->type = pairs.at("location");
        }
        catch(...) {
            ptr_->type = "unknown";
        }

        try {
            ptr_->acl = tokenize(pairs.at("acl"), ",:; ");
        }
        catch(...) {
            ptr_->acl = std::vector<std::string>();
        }

        try {
            ptr_->contact.display_ = pairs.at("display");
        }
        catch(...) {
            ptr_->contact.display_ = "Ext. " + ptr_->contact.user_;
        }
    }
}

auto registry::keys() const -> keyfile::keys {
    keyfile::keys tmp;

    if(ptr_) {
        const std::lock_guard lock(ptr_->mutex);
        tmp.insert(ptr_->keys.begin(), ptr_->keys.end());
    }
    return tmp;
}

auto registry::contact() const -> contact_t {
    if(ptr_) {
        const std::lock_guard lock(ptr_->mutex);
        return ptr_->contact;
    }
    return {};
}

auto registry::to(const event_t& event) const -> std::string {
    auto id = std::string_view(event.to()->url->username);
    auto name = event.to()->displayname;
    if(name && *name)
        return format("\"{}\" <{}>", name, local_alias(id));
    return format("<{}>", local_alias(id));
}

auto registry::from([[maybe_unused]] const event_t& event) const -> std::string {
    if(event) {
        auto id = std::string_view(event.from()->url->username);
        auto name = event.from()->displayname;
        auto group = std::string(event.from()->url->username);
        group[0] = static_cast<char>(std::toupper(group[0]));
        group += " Group";
        if(!name || !*name)
            name = const_cast<char *>(group.c_str());
        if(!registry::is_extension(id))
            return format("\"{}\" <{}>", name, local_alias(id));
    }
    auto name = contact().display();
    auto phone = ";user=phone";
    if(name.empty())
        return format("<{}{}>", local_uri(), phone);
    return format("\"{}\" <{}{}>", name, local_uri(), phone);
}

auto registry::origin() const -> contact_t {
    if(ptr_) {
        const std::lock_guard lock(ptr_->mutex);
        return ptr_->origin;
    }
    return {};
}

registry::operator bool() const {
    if(!context_)
        return false;
    time_t now{};
    time(&now);
    return expires() > now;
}

auto registry::operator!() const -> bool {
    if(!context_)
        return true;
    time_t now{};
    time(&now);
    return expires() <= now;
}

void registry::receipt(const event_t& event, time_t period) { // NOLINT
    auto ctx = event.context();
    auto tid = event.tid();
    auto msg = exosip::msg_t(nullptr);
    auto exp = std::to_string(period);
    auto receipt = R"({"url":")" + webserver +R"(","s":)" + std::to_string(reg_system->series) + R"(,"t":")" + shm()->token + R"(","n":")" + sitename + "\"}\n";

    const exosip::guard_t lock(ctx);
    eXosip_message_build_answer(ctx, tid, SIP_OK, &msg);
    osip_message_set_header(msg, "Expires", exp.c_str());
    if(period) {
        osip_message_set_content_type(msg, "application/json");
        osip_message_set_body(msg, receipt.c_str(), receipt.length());
    }
    eXosip_message_send_answer(ctx, tid, SIP_OK, msg);
}

auto registry::refresh(const event_t& event, time_t period) -> bool {
    if(!ptr_ || !event.request())
        return false;

    time_t now{};
    time(&now);

    auto registered = ptr_->expires > now;
    if(!registered) {
        auto request = event.request();
        context_ = event.context();
        ptr_->schema = event.schema();
        ptr_->agent = event.agent();

        auto allows = request->allows;

        shm_->flags.invite = shm_->flags.message = false;
        if(type() == "phone" || type() == "media")
            shm_->flags.invite = true;
        else if(type() == "chat")     // chat only endpoint
            shm_->flags.message = true;
        else if(osip_list_size(&allows)) {
            auto pos = 0;
            while(osip_list_eol(&allows, pos) == 0) {
                auto allowed = std::string_view(static_cast<osip_allow_t *>(osip_list_get(&allows, pos++))->value);
                if(allowed == "INVITE")
                    shm_->flags.invite = true;
                else if(allowed == "MESSAGE")
                    shm_->flags.message = true;
            }
        }
        else {
            shm_->flags.invite = shm_->flags.message = true;
        }
    }
    else if(context_ != event.context())
        return true;

    if(!event.count().empty())
        ptr_->count = util::get_count(event.count());

    ptr_->nonce = nonce_t(event.nonce());
    if(period) {
        now += period;
        ptr_->expires = now;
        // update origin in lock
        const std::lock_guard lock(ptr_->mutex);
        ptr_->origin.from(event);
    }
    else
        drop();

    return registered;
}

auto registry::verify(const event_t& event) -> bool {
    if(!ptr_) {
        logger.debug(5, "invalid user to verify");
        return false;
    }

    auto count = event.count();
    auto client = nonce_t(event.client());

    // if no nonce count, and none registered this session, okay...
    if(count.empty() && !ptr_->count)
        return true;

    logger.debug(5, "client nounce {}", count);
    auto cvalue = util::get_count(count);
    if(cvalue > ptr_->count) {
        ptr_->client = client.hex();
        ptr_->count = cvalue;
        return true;
    }

    if(cvalue < ptr_->count) {
        logger.debug(5,"failed counters: {} != {}", cvalue, ptr_->count);
        return false;
    }

    if(client.hex() != ptr_->client) {
        ptr_->client = client.hex();
        return true;
    }

    logger.debug(5, "failed client:{} == {}", client.hex(), ptr_->count);
    return false;
}

void registry::drop() {
    if(!ptr_ || !ptr_->expires)
        return;

    exec::hangup_user(*this, reason_t(SIP_REQUEST_TERMINATED, "user dropped"));
    logger.debug(5, "removing user {} from {}", value(), remote());
    clients.erase(remote());
    ptr_->expires = 0;
    ptr_->count = 0;
    shm_->flags.invite = shm_->flags.message = false;
    notify();
}

void registry::allow(exosip::msg_t msg) {
    auto allows = std::string("OPTIONS,REGISTER,INFO");

    if(shm_ && shm_->flags.message) {
        allows += ",MESSAGE";
    }
    if(shm_ && shm_->flags.invite) {
        allows += ",INVITE,BYE,CANCEL,ACK";
    }

    osip_message_set_header(msg, ACCEPT_ENCODING, "text/plain");
    osip_message_set_header(msg, ALLOW, allows.c_str());
    osip_message_set_header(msg, ACCEPT, "*/*");
}

void server::change_level(unsigned level) {
    reg_system->level = level;
    logger.set(level, logger_notify);
    logger.notice("logging level changed to {}", level);
}

void server::system_message(std::string_view text, std::string_view to, std::string_view subject) {
    struct pbx_msg msg{};
    memset(&msg, 0, sizeof(msg));
    msg.type = PBX_MESSAGE;
    str_copy(msg.body.message.to, sizeof(msg.body.message.to), to);
    str_copy(msg.body.message.text, sizeof(msg.body.message.text), text);
    if(!subject.empty())
        str_copy(msg.body.message.subj, sizeof(msg.body.message.subj), subject);
    exec::message_system(msg);
}

void server::exec_registry(const keyfile& config, int level) {
    using namespace std::chrono;
    using namespace process;

    auto args = args_t(config);
    auto shm = shm_open(SERVER_IPCPATH ".registry", O_RDWR | O_CREAT, 0660);
    if(shm < 0 || ftruncate(shm, mem_size) < 0)
        logger.crit(97, "failed to allocate registry");
    auto mem = map_t(shm, mem_size, true);
    ::close(shm);
    if(!is(mem))
        logger.crit(97, "failled to map registry");

    thread_id = std::this_thread::get_id();
    memset(mem.data(), 0, mem_size);
    reg_shared = static_cast<pbx_reg *>(mem.data());
    reg_system = reinterpret_cast<pbx_sys *>(&reg_shared[registry_entries]);

    key_storage.load("registry.keys");
    reg_system->version = PBX_VERSION;
    time(&reg_system->started);
    reg_system->first = dialplan_min;
    reg_system->level = level;
    reg_system->pid = getpid();
    reload(args);
    initial = false;

    logger.info("starting registry thread");
    while(!stopping) {
        sync_ptr msgs(msglist);
        if(!msgs->empty()) {
            auto copy = std::move(msgs->front());
            auto [actor, parms] = copy;
            msgs->pop_front();
            msgs.unlock();
            try {
                actor(parms);
            }
            catch(const std::exception& e) {
                logger.error("registry actor error: {}", e.what());
            }
            continue;
        }

        time_t latest = 0;
        if(timer) {
            msgs.unlock();
            latest = rescan();
            msgs.lock();
            if(!msgs->empty())
                continue;
            timer = false;
        }

        if(latest)
            timer = (cvar.wait_until(msgs, system_clock::from_time_t(latest)) == std::cv_status::timeout);
        else
            cvar.wait(msgs);
    }
    logger.info("stopping registry thread");
    sync_storage();
}

contact_t::contact_t(const osip_uri_t *uri) {
    if(!uri || !uri->host || uri->host[0] == 0)
        return;

    host_ = uri->host;
    if(uri->scheme && uri->scheme == "sips"s)
        port_ = 5061;
    else
        port_ = 5060;

    if(uri->port && uri->port[0] != 0)
        port_ = util::get_port(uri->port);

    if(uri->username)
        user_ = uri->username;
}

void contact_t::from(const event_t& event) {
    auto clist = event.request()->contacts;
    auto cport = uint16_t(5060);
    auto pos = 0;

    const osip_uri_t *url{nullptr};
    while(osip_list_eol(&clist, pos) == 0) {
        const auto *c = static_cast<osip_contact_t *>(osip_list_get(&clist, pos++));
        if(c && c->url && c->url->host) {
            url = c->url;
            // TODO: is ipv4 to be treated as a subset??
            /* if(ipv6 && net::is_ipv4(c->url->host))
                continue; */
            if(!ipv6 && net::is_ipv6(c->url->host))
                continue;
            break;
        }
    }

    if(url && url->username)
        user_ = url->username;
    if(url && url->host)
        host_ = url->host;
    if(url && url->port && url->port[0])
        cport = port_ = util::get_port(url->port);

    auto vlist = event.request()->vias;
    pos = 0;
    while(osip_list_eol(&vlist, pos) == 0) {
        auto via = static_cast<osip_via_t *>(osip_list_get(&vlist, pos++));
        auto rport = uint16_t(0);
        if(!via->host || !via->host[0])
            continue;
        if(via->port && via->port[0])
            cport = util::get_port(via->port);
        osip_uri_param_t *param{nullptr};
        osip_via_param_get_byname(via, const_cast<char *>("rport"), &param);
        if(param && param->gvalue && param->gvalue[0])
            rport = util::get_port(param->gvalue);

        // detect top nat...
        if(rport) {
            param = nullptr;
            osip_via_param_get_byname(via, const_cast<char *>("received"), &param);
            if(param && param->gvalue && param->gvalue[0]) {
                host_ = param->gvalue;
                port_ = rport;
                break;
            }
        }
        else {
            host_ = via->host;
            port_ = cport;
        }
    }

    if(host_[0] != '[' && host_.find(':') != std::string::npos)
        host_ = "[" + host_ + "]";
}

void exec::reload_registry(const keyfile& config) {
    dispatch(config, &reload);
}

void exec::stop_registry() {
    dispatch(args_t(), []([[maybe_unused]]const args_t& args){
        stopping = true;
        exec::stop_features();
        exec::stop_messages();
        exec::stop_calls();
    });
}

void exec::message_system(const pbx_msg& msg) {
    dispatch(std::make_shared<pbx_msg>(msg), [](const args_t& args){
        auto& ipc = *args.ipc;
        auto from = system_from();
        if(!ipc.body.message.text[0])
            return;

        ipc.body.message.to[sizeof(ipc.body.message.to) - 1] = 0;
        ipc.body.message.text[sizeof(ipc.body.message.text) - 1] = 0;
        ipc.body.message.subj[sizeof(ipc.body.message.subj) - 1] = 0;
        const auto *group = ipc.body.message.to;
        const std::string_view to = ipc.body.message.to;
        std::string_view text = ipc.body.message.text;
        std::string_view subject = ipc.body.message.subj;
        registry_map target;

        logger.debug(4, "system message for {}, text {}", to, text);
        if(to == "all")
            target = active_map;
        else if(group_map.count(group))
            target = group_map.at(group);
        else if(registry::is_extension(to))
            target = coverage_map[util::get_value(to) - dialplan_min];
        else
            return;

        target.apply([&](const auto& entry) {
            if(!entry.allow_message())
                return false;

            auto route = entry.route_to();
            auto context = entry.context();
            auto to = entry.origin_uri();
            auto msg = message(from, to, route, subject, text, "text/plain", context);
            exec::deliver_local(msg);
            return true;
        });
    });
}

void exec::pbx_action(const pbx_msg& ipc) {
    dispatch(std::make_shared<pbx_msg>(ipc), [](const args_t& args) {
        const auto ipc = *args.ipc;
        auto ext = ipc.body.registry.ext;
        if(ext < dialplan_min || ext >dialplan_max)
            return;

        auto id = std::to_string(ext);
        const char *key = ipc.body.registry.key;
        const char *value = ipc.body.registry.value;
        const auto& entry = mapped.at(ext - dialplan_min);
        auto shm = entry.shm();

        switch(ipc.type) {
        case PBX_ACTIVE:
            if(!is(entry))
                return;
            break;
        case PBX_CLEARS:
            key_storage.remove(id);
            key_updates = true;
            break;
        case PBX_ASSIGN:
            if(!key[0])
                return;
            key_storage[id][key] = value;
            key_updates = true;
            break;
        case PBX_PRESENCE:
            if(!is(entry))
                return;

            if(eq(value, "here"))
                shm->presence = pbx_reg::HERE;
            else if(eq(value, "away"))
                shm->presence = pbx_reg::AWAY;
            else if(eq(value, "dnd"))
                shm->presence = pbx_reg::DND;
            else if(eq(value, "gone"))
                shm->presence = pbx_reg::GONE;
            else
                return;
            break;
        default:
            return;
        }
    });
}

void exec::release_user(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto& entry = find(event.user());

        if(entry.empty()) {
            event.fail_message(SIP_NOT_FOUND);
            return;
        }

        const nonce_t nonce(event.nonce());
        if(nonce != entry.nonce()) {
            logger.debug(4, "release expected {}, got {}", entry.nonce().hex(), nonce.hex());
            event.fail_message(SIP_FORBIDDEN);  // TODO: challenge message??
            // this means we could not verify a fast exit, but we need to know
            // if the nonce was also valid from below us or not...
            return;
        }

        auto registered = is(entry);
        if(registered) {
            if(event.context() != entry.context()) {
                logger.warn("invalid context for user {}", event.user());
                event.fail_message(SIP_FORBIDDEN);
                return;
            }

            auto as = "user";
            auto agent = event.agent();
            if(agent.substr(0, 9) == "bordeaux/" && group_map.at("system")[entry.extension()]) {
                as = "feature server";
                if(event.user() == feature_server)
                    feature_server = "";
            }

            auto& shared = reg_shared[entry.extension() - dialplan_min];
            shared.activated = shared.expires = 0;
            logger.notice("releasing {} as {}", entry.origin_uri(), as);
            str_copy(shared.agent, sizeof(shared.agent), "-");
            str_copy(shared.name, sizeof(shared.name), "");
            str_copy(shared.id, sizeof(shared.id), "-");
            memset(shared.token, 0, sizeof(shared.token));
            shared.presence = pbx_reg::GONE;
            active_map -= entry.extension();
            presence.release(entry.extension());
            entry.drop();
            active--;
        }
        else
            logger.warn("cannot release {}", event.user());

        entry.receipt(event, 0);
    });
}

void exec::refresh_user(const event_t& event) {
    dispatch(event, [](const args_t& args){
        const auto& event = args.event;
        auto& entry = find(event.user());
        auto& shared = reg_shared[entry.extension() - dialplan_min];
        auto period = event.period();
        auto agent = event.agent();

        if(period < 60)
            period = 60;
        else if(period > max_expires)
            period = max_expires;

        if(entry.empty()) {
            event.fail_message(SIP_NOT_FOUND);
            return;
        }

        if(!is_authorized(entry, event)) {
            logger.error("failed to authorize {}", event.user());
            event.fail_message(SIP_FORBIDDEN);
            return;
        }

        auto prior_to = entry.remote();
        auto registered = entry.refresh(event, period);
        if(registered) {
            if(event.context() != entry.context()) {
                logger.warn("invalid context for user {}", event.user());
                event.fail_message(SIP_FORBIDDEN);
                return;
            }
            if(entry.remote() != prior_to) {
                logger.debug(5, "changing user {} from {} to {}", entry.value(), prior_to, entry.remote());
                clients.erase(prior_to);
                clients[entry.remote()] = entry;
            }
        }
        else {
            auto contact = entry.origin();
            auto pairs = entry.keys();
            auto stat = pairs.at("presence");

            active_map += entry.extension();
            net::store(shared.address, contact.host(), contact.port());
            str_copy(shared.agent, sizeof(shared.agent), agent);
            str_copy(shared.name, sizeof(shared.name), entry.display());
            str_copy(shared.id, sizeof(shared.id), entry.user_id());

            if(stat == "dnd")
                shared.presence = pbx_reg::DND;
            else if(stat == "gone" || stat == "off" || stat == "closed")
                shared.presence = pbx_reg::GONE;
            else if(stat == "away")
                shared.presence = pbx_reg::AWAY;
            else
                shared.presence = pbx_reg::HERE;

            auto as = "user";
            if(agent.substr(0, 9) == "bordeaux/" && group_map.at("system")[entry.extension()]) {
                as = "feature server";
                feature_server = event.user();
            }

            uint8_t random[16];
            crypto::rand(random, sizeof(random));
            const auto token = std::to_string(entry.extension()) + ":" + to_hex(random, sizeof(random));
            str_copy(shared.token, sizeof(shared.token), token);

            time(&shared.activated);
            logger.notice("activating {} as {}", entry.origin_uri(), as);
            active++;
            notify();
            clients[entry.remote()] = entry;
            if(entry.welcome())
                system_message(server_messages.at("welcome"));
        }
        logger.debug(4, "refreshing {} using {} for {} seconds", entry.origin_uri(), event.algorithm(), period);
        shared.expires = entry.expires();
        entry.receipt(event, period);
    });
}

void exec::option_user(const event_t& event) {
    dispatch(event, [](const args_t& args){
        const auto& event = args.event;
        auto to = std::string_view(event.to()->url->username);
        auto context = event.context();
        auto tid = event.tid();
        auto msg = exosip::msg_t(nullptr);
        auto code = SIP_OK;
        auto entry = registry();
        auto group = std::string{to};
        std::list<std::string> contacts;

        if(to == "system")
            contacts.emplace_front(sitefrom);
        else if(group_map.count(group)) {
            auto target = group_map.at(group);
            auto count = target.apply([&contacts](const auto& user) {
                contacts.emplace_back(user.local_uri());
                return true;
            });
            if(!count)
                code = SIP_TEMPORARILY_UNAVAILABLE;
        }
        else if(registry::is_extension(to)) {
            entry = mapped[util::get_value(to) - dialplan_min];
            if(is(entry))
                contacts.emplace_front(entry.local_uri());
            else
                code = SIP_TEMPORARILY_UNAVAILABLE;
        }
        else
            code = SIP_NOT_FOUND;

        const exosip::guard_t lock(context);
        eXosip_options_build_answer(context, tid, code, &msg);
        if(!msg)
            return;

        for(const auto& contact : contacts)
            osip_message_set_header(msg, CONTACT, contact.c_str());

        if(is(entry))
            entry.allow(msg);
        else
            allow_local(msg);
        eXosip_options_send_answer(context, tid, code, msg);
    });
}

void exec::publish_user(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto user = find(event.user());

        if(user.empty()) {
            event.fail_message(SIP_NOT_FOUND);
            return;
        }

        if(!is_authorized(user, event)) {
            logger.error("failed to authorize {}", event.user());
            event.fail_message(SIP_FORBIDDEN);
            return;
        }

        auto to = event.to();
        auto tid = event.tid();
        auto type = event.content();
        auto target = "";
        auto entry = registry();
        auto publish = event.publish();
        auto expires = event.period();
        auto context = event.context();

        if(to && to->url && to->url->username) {
            target = to->url->username;
            entry = find(target);
        }

        if(!event.body()) {
            type = "expired";
        }
        else if(type == "application/pidf+xml" && is(entry)) {
            auto body = content_view(event.body());
            auto shm = entry.shm();
            auto activity = (body.find("activities>") != std::string::npos);
            if(body.find("basic>open</") != std::string::npos)
                shm->presence = activity ? pbx_reg::AWAY : pbx_reg::HERE;
            else if(body.find("basic>closed</") != std::string::npos)
                shm->presence = activity ? pbx_reg::DND : pbx_reg::GONE;
        }

        if(publish == "presence") {
            eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
            publish_event(presence, entry, event);
        }
        else {
            logger.error("publish unknown {} from {}", publish, target);
            event.fail_message(SIP_NOT_ACCEPTABLE_HERE);
            return;
        }

        if(expires)
            logger.debug(4, "publish {} {} from {} for {}s", publish, type, target, expires);
        else
            logger.debug(4, "publish {} from {} released", publish, target);
    });
}

void exec::message_user(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto user = find(event.user());
        auto system_user = group_map.at("system")[user.extension()];

        if(user.empty()) {
            event.fail_message(SIP_NOT_FOUND);
            return;
        }

        if(!is_authorized(user, event)) {
            logger.error("failed to authorize {}", event.user());
            event.fail_message(SIP_FORBIDDEN);
            return;
        }

        registry_map target;
        auto context = event.context();
        auto body = event.body();
        auto type = event.content();
        auto text = content_view(body);
        auto tid = event.tid();
        auto to = std::string_view(event.to()->url->username);
        auto group = event.to()->url->username;
        auto from = user.from(event);
        auto subject = event.subject();

        logger.debug(4, "message from \"{}\" for {}, text {}", from, to, text);

        if(to == "system") {
            while(!text.empty() && text.front() == ' ')
                text.remove_prefix(1);

            auto sysop = group_map.at("sysop")[user.extension()] || system_user;
            if(sysop && text.length() > 6 && text.substr(0, 6) == "/sysop") {
                cli_sysop(user, text);
            }
            else if(!system_user && (text.empty() || text.front() != '/' || !cli_command(user, text)))
                cli_help(user);

            const exosip::guard_t lock(context);
            eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
            return;
        }

        if(system_user) {
            system_message(text, to, subject);
            const exosip::guard_t lock(context);
            eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
            return;
        }

        if(group_map.count(group))
            target = group_map.at(group);
        else if(registry::is_extension(to))
            target = coverage_map[util::get_value(to) - dialplan_min];
        else {
            event.fail_message(SIP_NOT_FOUND);
            return;
        }

        auto count = target.apply([&](const auto& entry) {
            if(!entry.allow_message())
                return false;

            auto route = entry.route_to();
            auto context = entry.context();
            auto to = entry.to(event);
            auto msg = message(from, to, route, subject, text, type, context);
            exec::deliver_local(msg);
            return true;
        });

        if(count)
            eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
        else
            event.fail_message(SIP_TEMPORARILY_UNAVAILABLE);
    });
}

void exec::invite_user(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto user = find(event.user());

        if(!is_authorized(user, event)) {
            logger.error("failed to authorize {}", event.user());
            event.fail_call(SIP_FORBIDDEN);
            return;
        }

        registry_map invites, target, delayed;
        auto to = std::string_view(event.to()->url->username);
        //auto from = user.local_uri();
        auto group = event.to()->url->username;
        auto subject = event.subject();
        auto display = user.contact().display();
        //auto body = event.body();
        //auto sdp = content_view(body);
        auto busy = false;
        auto up = user.shm();

        // Check pre-conditions...
        if(group_map.count(group))
            invites = group_map.at(group);
        else if(registry::is_extension(to)) {
            auto map = util::get_value(to) - dialplan_min;
            invites = coverage_map[map];
            delayed = delayed_map[map];
        }

        // If nothing to invite, bad to...
        if(invites.empty()) {
            event.fail_call(SIP_NOT_FOUND);
            return;
        }

        // Make sure we have capacity...
        if(++up->count > up->lines) {
            --up->count;
            event.fail_call(SIP_BUSY_HERE);
            return;
        }

        logger.debug(4, "call from \"{}\" for {}, subject {}", display, to, subject);

        // Attempt target calling...
        invites.apply([&](const auto& entry) {
            if(target[entry.extension()])
                return false;

            if(!entry.allow_invite())
                return false;

            // maybe forward busy support...etc...
            auto ep = entry.shm();
            if(++ep->count <=  ep->lines) {
                target += entry.extension();
                return true;
            }
            busy = true;
            --ep->count;
            return false;
        });

        // If invite targets, define for start ringing...
        if(!target.empty() || !delayed.empty()) {

            auto call_id = next_callid();
            if(call_id) {
                exec::create_call(call_id, event, user);
                target.apply([&call_id](const auto& to) {
                    exec::invite_call(call_id, to);
                    return true;
                });
                if(!delayed.empty()) {
                    delayed ^= target;
                    delayed.apply([&call_id](const auto& entry) {
                        if(!entry.allow_coverage())
                            return false;
                        exec::invite_delayed(call_id, entry);
                        return true;
                    });
                }
                exec::start_call(call_id, event, user);
                return;
            }

            // if no call, clean up ring requests...
            target.apply([](const auto& entry) {
                --entry.shm()->count;
                return true;
            });
            logger.debug(4, "call busy everywhere for {}", display);
            event.fail_call(SIP_BUSY_EVRYWHERE);
            return;
        }

        // If no targets or no call, clean up invite and exit...
        --up->count;
        logger.debug(4, "call not possible for {}", display);
        if(busy)
            event.fail_call(SIP_BUSY_HERE);
        else
            event.fail_call(SIP_TEMPORARILY_UNAVAILABLE);
    });
}

void exec::feature_code(const event_t& event) {
    dispatch(event, [](const args_t& args) {
        const auto& event = args.event;
        auto& user = find(event.user());

        // system group cannot do feature dialing, redundent...
        if(group_map.at("system")[user.extension()]) {
            event.fail_call(SIP_NOT_FOUND);
            return;
        }

        if(!is_authorized(user, event)) {
            logger.error("failed to authorize {}", event.user());
            event.fail_call(SIP_FORBIDDEN);
            return;
        }

        auto feature = std::make_shared<feature_data>();
        feature->code = event.to()->url->username;
        feature->server = find(feature_server);
        feature->origin = user;
        feature->sysop = group_map.at("sysop")[user.extension()];
        exec::dialed_feature(feature, event);
    });
}

