/*
 * Copyright (C) 2022 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"       // IWYU pragma: keep
#include "ipc.hpp"

#include <moderncli/process.hpp>
#include <moderncli/keyfile.hpp>
#include <moderncli/args.hpp>

#include <sys/stat.h>

namespace {
args::option opt_config('c',"--config",     "use config file", "path");
args::option helpflag('h',  "--help",       "display this list");
args::option althelp('?',   nullptr, nullptr);
} // end namespace

auto main(int argc, const char **argv) -> int {
    using namespace process;

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({"pbx-realm [options]"}, {
                "pbx-realm " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Realm of pbx services"
            });
            return 0;
        }

        if(is(opt_config)) {
            keyfile config;
            config.load(*opt_config);
            auto& hostname = config["server"]["hostname"];
            auto& realm = config["server"]["realm"];

            if(hostname.empty()) {
                char buf[256];
                auto result = gethostname(buf, sizeof(buf));
                if(result != 0)
                    throw std::runtime_error("cannot get hostname");
                buf[sizeof(buf) - 1] = 0;
                hostname = buf;
            };
            //net_fqdn(hostname);
            if(realm.empty())
                realm = hostname;
            print("{}\n", realm);
            return 0;
        }

        auto rshm = shm_open(SERVER_IPCPATH ".registry", O_RDONLY, 0660);
        if(rshm < 0)
            throw std::runtime_error("failed");
        struct stat rino{};
        if(fstat(rshm, &rino) < 0)
            throw std::runtime_error("offline");
        auto rmax = (rino.st_size - sizeof(pbx_sys)) / sizeof(pbx_reg);
        auto rmem = map_t(rshm, rino.st_size, false);
        if(!is(rmem))
            throw std::runtime_error("failed");
        auto reg = static_cast<pbx_reg*>(rmem.data());
        auto sys = reinterpret_cast<pbx_sys *>(&reg[rmax]);
        print("{}\n", sys->realm);
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-realm [options]\nerror: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        die(-1, "error: unknown error\n");
    }
    return 0;
}
