/*
 * Copyright (C) 2021 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"       // IWYU pragma: keep
#include "ipc.hpp"

#include <moderncli/args.hpp>
#include <moderncli/process.hpp>

#include <mqueue.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>

using namespace std::literals::string_literals;

namespace {
args::option agentid('a',   "--agent",      "show agent in status");
args::option callflag('c',  "--calls",      "show active calls");
args::option display('d',   "--display",    "show display name in status");
args::option helpflag('h',  "--help",       "display this list");
args::option loglevel('l',  "--level",      "set log level", "0-9");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option registry('r',  "--registry",   "show registry status");
args::option token('t',     "--token",      "show url bearer token");
args::option reload(0,      "--reload",     "reload server config");
args::option althelp('?',   nullptr, nullptr);

class ipc_error final : public std::exception {
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

auto scount(const char *cp, size_t max = 256) {
    size_t count = 0;
    while(cp && *cp && count < max) {
        ++count;
        ++cp;
    }
    return count;
}

auto uptime(time_t when) {
    auto now = time(nullptr) - when;
    if(now < 90)
        return format("{}s", now);
    now /= 60;
    if(now < 90)
        return format("{}m", now);
    now /= 60;
    if(now < 96)
        return format("{}h", now);
    now /= 24;
    if(now < 28)
        return format("{}d", now);
    return format("{}w", now / 7);
}
} // end namespace

auto main(int argc, const char **argv) -> int {
    using namespace process;

    auto exit_code = 0;
    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({"pbx-status [options]"}, {
                "pbx-status " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Send command to server ipc"
            });
            return 0;
        }

        pbx_msg msg{.type = PBX_ACTIVE, .ver = PBX_VERSION};
        if(is(loglevel)) {
            msg.type = PBX_LEVEL;
            msg.body.level = std::stoi(*loglevel);
        }
        else if(is(reload)) {
            msg.type = PBX_RELOAD;
            msg.body.message.text[0] = 0;
        }

        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq == MQD_INVALID)
            throw ipc_error("mqueue failed", 4);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);

        auto calls = 0U;
        auto users = 0U;
        auto busy = 0U;
        auto rmax = 80U;
        auto cshm = shm_open(SERVER_IPCPATH ".calls", O_RDONLY, 0660);
        if(cshm < 0)
            throw ipc_error("call access failed", 4);
        struct stat cino{};
        if(fstat(cshm, &cino) < 0)
            throw ipc_error("stat failed", 4);
        auto cmax = cino.st_size / sizeof(pbx_call);
        auto cmem = map_t(cshm, cino.st_size, false);
        if(!is(cmem))
            throw ipc_error("call mapping failed", 4);

        auto rshm = shm_open(SERVER_IPCPATH ".registry", O_RDONLY, 0660);
        if(rshm < 0)
            throw ipc_error("registry access failed", 4);
        struct stat rino{};
        if(fstat(rshm, &rino) < 0)
            throw ipc_error("stat failed", 4);
        rmax = (rino.st_size - sizeof(pbx_sys)) / sizeof(pbx_reg);
        auto rmem = map_t(rshm, rino.st_size, false);
        if(!is(rmem))
            throw ipc_error("registry mapping failed", 4);

        auto call = static_cast<pbx_call *>(cmem.data());
        auto reg = static_cast<pbx_reg*>(rmem.data());
        auto sys = reinterpret_cast<pbx_sys *>(&reg[rmax]);
        auto trunk = &sys->trunk;

        for(auto i = 0U; i < cmax; ++i) {
            const auto& entry = call[i];
            if(entry.id > 0)
                ++calls;
            if(is(callflag) && entry.id > 0) {
                auto status{"unknown"};
                switch (entry.state) {
                case pbx_call::INVITING:
                    status = "inviting";
                    break;
                case pbx_call::RINGING:
                    status = "ringing";
                    break;
                case pbx_call::CLOSING:
                    status = "closing";
                    break;
                case pbx_call::CONNECTED:
                    status = "active";
                    break;
                }

                auto type{"unknown"};
                switch(entry.type) {
                case pbx_call::LOCAL:
                    type = "local";
                    break;
                case pbx_call::INBOUND:
                    type = "inbound";
                    break;
                case pbx_call::OUTGOING:
                    type = "outgoing";
                    break;
                case pbx_call::NONE:
                    type = "none";
                    break;
                }

                time_t duration{};
                time(&duration);
                duration -= entry.created;

                print("0x{:08x} {} {}, {} participant(s), {}s duration\n", entry.id, status, type, entry.ringing ? entry.ringing : entry.active, duration);
            }
        }

        for(auto i = 0U; i < rmax; ++i) {
            const auto& entry = reg[i];
            time_t expires{};
            time(&expires);
            if(!entry.activated || expires > entry.expires)
                continue;

            auto flags = "--"s;
            if(entry.flags.invite)
                flags[0] = 'i';
            if(entry.flags.message)
                flags[1] = 'm';

            auto presence = "here";
            if(entry.count >= entry.lines)
                presence = "busy";
            else if(entry.count)
                presence = "call";
            switch(entry.presence) {
            case pbx_reg::GONE:
                presence = "gone";
                break;
            case pbx_reg::DND:
                presence = "dnd ";
                break;
            case pbx_reg::AWAY:
                presence = "away";
                break;
            default:
                break;
            }
            if(is(registry) || is(agentid) || is(display) || is(token)) {
                std::string endpoint;
                if(is(agentid))
                    endpoint = std::string(entry.agent);
                else if(is(display))
                    endpoint = std::string(entry.name);
                else if(is(token))
                    endpoint = std::string(entry.token);
                else {
                    auto addr = reinterpret_cast<const struct sockaddr *>(&entry.address);
                    char host[128];
                    uint16_t port = 0;
                    memset(host, 0, sizeof(host));
                    switch(addr->sa_family) {
                    case AF_INET:
                        getnameinfo(addr, sizeof(struct sockaddr_in), host, sizeof(host), nullptr, 0, NI_NUMERICHOST);
                        port = ntohs(reinterpret_cast<const struct sockaddr_in*>(addr)->sin_port);
                        break;
                    case AF_INET6:
                        getnameinfo(addr, sizeof(struct sockaddr_in6), host + 1, sizeof(host) - 2, nullptr, 0, NI_NUMERICHOST);
                        host[0] = '[';
                        host[scount(host)] = ']';
                        port = ntohs(reinterpret_cast<const struct sockaddr_in6*>(addr)->sin6_port);
                        break;
                    default:
                        snprintf(host, sizeof(host), "unknown");
                    }
                    endpoint = format("{}:{}", host, port);
                }
                auto used = entry.count.load();
                if(used > entry.lines)
                    used = entry.lines;
                print("{:<4}{:32} {:4} {} {:8} {}/{}\n", i + sys->first, endpoint, presence, flags, uptime(entry.activated), used, entry.lines);
            }
            if(entry.count)
                ++busy;
            ++users;
        }
        if(!is(quiet)) {
            auto remotes = trunk->count.load();
            if(remotes > trunk->lines)
                remotes = trunk->lines;
            print("online={}/{} busy={} calls={}/{} remotes={}/{} uptime={} log={}\n", users, rmax, busy, calls, cmax, remotes, trunk->lines, uptime(sys->started), sys->level);
        }

    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-status [options]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}
